//---------------------------------------------------------------------------

#pragma hdrstop

#define _EXPORTING
#include "..\miniscope\device\MiniscopeDevice.h"
#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include <assert.h>
#include "ArmScope.h"
#include <algorithm>	// needed by Utils::in_group
#include "Utils.h"
#include <string>
#include <fstream> 
#include <json/json.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)

static const struct S_DEVICE_DLL_INTERFACE dll_interface =
	{DLL_INTERFACE_MAJOR_VERSION, DLL_INTERFACE_MINOR_VERSION};

// callback ptrs
CALLBACK_LOG lpLogFn = NULL;
CALLBACK_CONNECT lpConnectFn = NULL;
CALLBACK_DATA lpDataFn = NULL;
CALLBACK_TRIGGER lpTriggerFn = NULL;


void *callbackCookie;	///< used by upper class to distinguish library instances when receiving callbacks

void __stdcall GetMiniscopeInterfaceDescription(struct S_DEVICE_DLL_INTERFACE* interf)
{
	interf->majorVersion = dll_interface.majorVersion;
	interf->minorVersion = dll_interface.minorVersion;
}

void Log(char* txt)
{
	if (lpLogFn)
		lpLogFn(callbackCookie, txt);
}

void Data(short* data, int size)
{
	if (lpDataFn)
		lpDataFn(callbackCookie, data, size);
}

void Connect(int state, char *szMsgText)
{
	if (lpConnectFn)
		lpConnectFn(callbackCookie, state, szMsgText);
}

void Trigger(void)
{
	if (lpTriggerFn)
		lpTriggerFn(callbackCookie);
}

void __stdcall SetCallbacks(void *cookie, CALLBACK_LOG lpLog, CALLBACK_CONNECT lpConnect,
	CALLBACK_DATA lpData, CALLBACK_TRIGGER lpTrigger)
{
	assert(cookie && lpLog && lpConnect && lpData && lpTrigger);
	lpLogFn = lpLog;
	lpConnectFn = lpConnect;
	lpDataFn = lpData;
	lpTriggerFn = lpTrigger;
	callbackCookie = cookie;
	lpLogFn(callbackCookie, "armscope_dma.dll loaded. Yay!");

	armScope.callbackLog = Log;
	armScope.callbackData = Data;
	armScope.callbackConnect = Connect;
	armScope.callbackTrigger = Trigger;
}

static enum E_TRIGGER_TYPE eTrigger[] = { TRIGGER_MANUAL, TRIGGER_SLOPE_UP, TRIGGER_SLOPE_DOWN, TRIGGER_CONTINUOUS };
static enum E_COUPLING_TYPE eCoupling[] = { COUPLING_DC, COUPLING_AC };
/** \sa size of Armscope::output_buffer when increasing max size
*/
static unsigned int iBuf[] = { 4096, 8192, 12288, 16384, 20480, 32768 };
static float fSens[] = { 0.000547, 0.001602, 0.004766, 0.014335, 0.042969, 0.128906 };
const float fBasicSampl = 3.9646768e-6;	// not a round number unfortunately
static float fSampl[] = { fBasicSampl*1024, fBasicSampl*512, fBasicSampl*256, fBasicSampl*128, fBasicSampl*64, fBasicSampl*32,
	fBasicSampl*16, fBasicSampl*8, fBasicSampl*4, fBasicSampl*2, fBasicSampl };

unsigned int iSensitivity[2];
E_COUPLING_TYPE eCouplingSettings[2];
bool bChannelEnable[2];

/** \brief Calibration data for single sensitivity range
*/
static struct S_CALIBRATION
{
	int offset;
	float gain;
} calibration[sizeof(fSens)/sizeof(fSens[0])];

void __stdcall GetDeviceCapabilities(struct S_SCOPE_CAPABILITIES **caps)
{
	static struct S_SCOPE_CAPABILITIES capabilities =
	{
        2,		//channels
		8,		//bits/sample
		sizeof(fSens)/sizeof(fSens[0]), fSens,
		sizeof(fSampl)/sizeof(fSampl[0]), fSampl,
		sizeof(eTrigger)/sizeof(eTrigger[0]), eTrigger,
		sizeof(eCoupling)/sizeof(eCoupling[0]), eCoupling,
		sizeof(iBuf)/sizeof(iBuf[0]), iBuf
	};
	*caps = &capabilities;
}

void __stdcall ShowSettings(HANDLE parent)
{
	MessageBox(parent, "Armscope.dll has no additional settings.", "Device DLL", MB_ICONINFORMATION);
}

int __stdcall SetSensitivity(int iChannelId, int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	return armScope.SetSensitivity(iChannelId, iId);
}

int __stdcall SetSamplingPeriod(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(fSampl)/sizeof(fSampl[0])))
		return -1;
	return armScope.SetSampling(sizeof(fSampl)/sizeof(fSampl[0]) - iId - 1);
}

int __stdcall SetTriggerSource(int iChannelId, int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eTrigger)/sizeof(enum E_TRIGGER_TYPE)))
		return -1;
	return armScope.SetTriggerType(iChannelId, eTrigger[iId]);
}

int __stdcall SetTriggerLevel(int level)
{
	return armScope.SetTriggerLevel(level);	
}

int __stdcall SetAcqMode(int iId)
{
	return 0;
}

int __stdcall SetCouplingType(int iChannelId, int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eCoupling)/sizeof(eCoupling[0])))
		return -1;
	return armScope.SetCouplingType(iChannelId, eCoupling[iId]);
}

int __stdcall SetBufferSize(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(iBuf)/sizeof(unsigned int)))
		return -1;
	return armScope.SetBufferSize(iBuf[iId]);
}

int __stdcall SetOffset(float fOffset)
{
	return 0;
}

int __stdcall Connect(void)
{
	return armScope.Connect();
}

int __stdcall Disconnect(void)
{
	return armScope.Disconnect();
}

int __stdcall AcquireStart(void)
{
	armScope.SetTriggerType(0, TRIGGER_CONTINUOUS);
	return armScope.ManualTrigger();
}

int __stdcall AcquireStop(void)
{
	return armScope.SetTriggerType(0, TRIGGER_MANUAL);
}

int __stdcall ManualTrigger(void)
{
	return armScope.ManualTrigger();
}

static bool bSettingsReaded = false;

static int GetDefaultSettings(struct S_SCOPE_SETTINGS* settings)
{
	for (unsigned int i=0; i<sizeof(iSensitivity)/sizeof(iSensitivity[0]); i++)
	{
		iSensitivity[i] = 0;
		bChannelEnable[i] = true;
	}
	settings->iChannelCount = 2;
	settings->pChannelEnable = bChannelEnable;
	settings->pSensitivity = iSensitivity;
	settings->iSamplingPeriod = 0;
	settings->eTriggerType = TRIGGER_MANUAL;
	settings->iTriggerLevel = 0;
	settings->iTriggerSrcChannel = 0;
	for (unsigned int i=0; i<sizeof(iSensitivity)/sizeof(iSensitivity[0]); i++)
	{
		eCouplingSettings[i] = COUPLING_DC;
	}
	settings->eCouplingType = eCouplingSettings;
	settings->iBufferSize = 4096;
	for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		calibration[i].offset = 0;
		calibration[i].gain = 1.0;
	}
	bSettingsReaded = true;
	return 0;
}

int __stdcall GetMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return GetDefaultSettings(settings);

	settings->iTriggerSrcChannel = 0;

	Json::Value root;   // will contains the root value after parsing.
	Json::Reader reader;

	std::ifstream ifs(path.c_str());
	std::string strConfig((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
	ifs.close();

	bool parsingSuccessful = reader.parse( strConfig, root );
	if ( !parsingSuccessful )
		return GetDefaultSettings(settings);

	settings->pChannelEnable = bChannelEnable;

	const Json::Value sens = root["Sensitivity"];
	unsigned int cnt = std::min(sizeof(iSensitivity)/sizeof(iSensitivity[0]), sens.size());
	for (unsigned int i=0; i<cnt; i++)
	{
		iSensitivity[i] = sens.get(i, 0).asInt();
		if (iSensitivity[i] >= sizeof(fSens)/sizeof(fSens[0]))
			iSensitivity[i] = 0;
	}
	for (unsigned int i=cnt; i < sizeof(iSensitivity)/sizeof(iSensitivity[0]); i++)
	{
		iSensitivity[i] = 0;
	}
	settings->pSensitivity = iSensitivity;

	settings->iSamplingPeriod = root.get("SamplingPeriod", 0).asInt();
	if (settings->iSamplingPeriod >= sizeof(fSampl)/sizeof(fSampl[0]))
		settings->iSamplingPeriod = 0;

	int trigger = root.get("TriggerType", TRIGGER_MANUAL).asInt();
	if (Utils::in_group(trigger, eTrigger))
		settings->eTriggerType = (E_TRIGGER_TYPE)trigger;
	else
		settings->eTriggerType = TRIGGER_MANUAL;

	settings->iTriggerLevel = root.get("TriggerLevel", 0).asInt();
	if (settings->iTriggerLevel < -100 || settings->iTriggerLevel > 100)
		settings->iTriggerLevel = 0;

	const Json::Value coupl = root["CouplingType"];
	settings->eCouplingType = eCouplingSettings;
	cnt = std::min(sizeof(eCouplingSettings)/sizeof(eCouplingSettings[0]), coupl.size());
	for (unsigned int i=0; i<cnt; i++)
	{
		int coupling = coupl.get(i, COUPLING_DC).asInt();
		if (Utils::in_group(coupling, eCoupling))
			settings->eCouplingType[i] = (E_COUPLING_TYPE)coupling;
		else
			settings->eCouplingType[i] = COUPLING_DC;
	}
	for (unsigned int i=cnt; i < sizeof(eCouplingSettings)/sizeof(eCouplingSettings[0]); i++)
	{
		settings->eCouplingType[i] = COUPLING_DC;
	}

	settings->iBufferSize = root.get("BufferSize", 4096).asInt();
	if (!Utils::in_group(settings->iBufferSize, iBuf))
		settings->iBufferSize = iBuf[0];

	// read calibration data also
	const Json::Value rootcal = root["Calibration"];
	cnt = std::min(sizeof(calibration)/sizeof(calibration[0]), rootcal.size());
	for (unsigned int i=0; i<cnt; i++)
	{
		calibration[i].offset = rootcal[i].get("Offset", 0).asInt();
		calibration[i].gain = rootcal[i].get("Gain", 1.0).asDouble();
	}
	for (unsigned int i=cnt; i < sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		calibration[i].offset = 0;
		calibration[i].gain = 1.0;
	}
	bSettingsReaded = true;
	return 0;
}

int __stdcall SaveMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
	Json::Value root;
	Json::StyledWriter writer;
	for (unsigned int i=0; i<sizeof(iSensitivity)/sizeof(iSensitivity[0]); i++)
	{
		  root["Sensitivity"][i] = iSensitivity[i];
	}
	root["SamplingPeriod"] = settings->iSamplingPeriod;
	root["TriggerType"] = settings->eTriggerType;
	root["TriggerLevel"] = settings->iTriggerLevel;
	for (unsigned int i=0; i<sizeof(eCouplingSettings)/sizeof(eCouplingSettings[0]); i++)
	{
		  root["CouplingType"][i] = eCouplingSettings[i];
	}
	root["BufferSize"] = settings->iBufferSize;

	// add calibration data also
	const Json::Value rootcal = root["Calibration"];
	for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		  root["Calibration"][i]["Offset"] = calibration[i].offset;
		  root["Calibration"][i]["Gain"] = calibration[i].gain;
	}

	std::string outputConfig = writer.write( root );

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return -1;

    std::ofstream ofs(path.c_str());
    ofs << outputConfig;
	ofs.close();

	return 0;
}

int __stdcall StoreCalibration(int range, int offset, float gain)
{
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	calibration[range].offset = offset;
	calibration[range].gain = gain;
	return 0;
}

int __stdcall GetCalibration(int range, int* const offset, float* const gain)
{
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	struct S_SCOPE_SETTINGS settings;
	if (bSettingsReaded == false)
	{
		int status = GetMiniscopeSettings(&settings);
		if (status)
			return status;
	}
	*offset = calibration[range].offset;
	*gain = calibration[range].gain;
	return 0;
}

